// TREEMM: a software for identification of bipartite motifs in promoter sequences
// Copyright (C) 2011  Pierre NICOLAS
//
// This file is part of TREEMM.
//
//     TREEMM is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     Foobar is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with TREEMM.  If not, see <http://www.gnu.org/licenses/>.

#ifndef MMODEL_H
#define MMODEL_H

#include <string>
#include <gsl/gsl_rng.h>
using namespace std;

#include "Sequences.h"
#include "Tree.h"


class Mmodel
{
private:
  
  gsl_rng* _r;
  Tree* _tree;

  int _maxdistfrom3end;
  // model parameters

  bool _allownomotif;
  int _rbackg;
  int _nmotifs;
  int _dimlikcomput;
  int _lmaxspacer;
  int _lminspacer;
  // spacer = distance between 
  //the leftmost pos of box1 and the rightmost pos of box2 
  int _lmaxbox1;
  int _lmaxbox2;
  int _lminbox1;
  int _lminbox2;
  int _lmaxvarspacer; 
  int _lminvarspacer; 
  int* _lvarspacer; 

  double _alpha_dirich_nuc;
  double _logltot;

  double* _pmotifs;
  double* _pmotifs_ontree;
  double _alpha_evolontree; // rate of motif change along the branches of the tree, the proportion of the each motifs is given by _pmotifs_ontree
  double _epsilon_evolontree; // rate of motif alloc independent of the tree, the proportion of the each motifs is given by _pmotifs_ontree

  double* _p_pos_motif;  //for each pos in the seq give the prior proba of motif occurence 
  int* _breakpoints_p_pos_motif; // 1 if breakpoint, 0 otherwise
  int _nbins_p_pos_motif;
  int* _count_pos_motif; // 1 if breakpoint, 0 otherwise

  int* _lspacer;
  double** _pvarspacer;
  int* _lbox1; // box1 is downstream
  int* _lbox2;
  int* _ref_posbox1;// the position within the box1 that serves as the ref
  double*** _log_pmatbox1; // _log_pmatbox1[imotif][ipos][inuc]
  double*** _log_pmatbox2;
  double* _pbox2; // the probability of the second (-35) box
  double _minpbox2; // the probability of the second (-35) box
  int* _box2isoptional; // the probability of the second (-35) box

  double* _log_backgroundM1;
  double* _log_backgroundMr;

  int* _track_posbox1;// the position of box1 across iterations. From one iteration to another: +1 if the motif is shortened on the left, -1 if the motif is lengthened. This allows to align the motifs across iteration. 
  int* _track_posbox2;

  // allocation variables (size depends on the nb of sequences)
  int _nseqs;
  int _lseq;
  bool _box2oblig;

  int* _alloc_imotif; // id of the motif for each seq
  int* _alloc_tbox1refpos; //the position of the ref pos of box1 in each seq
  int* _alloc_varspacer; // set to -1 if no box2
  
  // likelihood computation (size depends on the nb and the length of sequences)

  double** _logl_cumsum_background_seq; // _logl_cumsum_background_seq[iseq][t]
  double*** _logl_ratio_box1; //  _logl_ratio_box1[iseq][imotif][t]
  // Here "t" refers to the leftmost position of box1
  double*** _logl_ratio_box2; 
  // Here "t" refers to the rightmost position of box2
  double** _logl_seq_by_motif; // _logl_seq_by_motif[iseq][imotif]
  double*** _ratio_seqt; // _logl_ratio_seqt[iseq][imotif][t]

  double** _tmp_ratio_seqt; // _logl_ratio_seqt[iseq][t]

  // useful
  int*** _count_nuc_matbox1;
  int*** _count_nuc_matbox2;
  int** _count_varspacer;
  int** _count_box2nobox2;
  int* _count_nuc_backg;

  int**** _count_nuc_matbox1_subtree; // [inode][submotif(0 or 1)][ipos][inuc]
  int**** _count_nuc_matbox2_subtree; // [inode][submotif(0 or 1)][ipos][inuc]
  int** _count_varspacer_subtree; // [inode][ivs]
  int** _count_box2nobox2_subtree; // [inode][0 or 1]
  int* _same_count_as_subtree; // [inode]
  double* _liktreesplit_subtree; // [inode]
  bool* _allowedsplit_subtree;
  double** _loglikbackground_subtree; // [inode][submotif(0 or 1)]
			 
  double _invtemp;

  void computeBackgroundM1(Sequences* seqs);
  void computeBackgroundMr(Sequences* seqs);

public:
  Mmodel(int nmotifs, bool allownomotif, double alpha_dirich_nuc, int rbackg, int lminspacer, int lmaxspacer, int lminvarspacer, int lmaxvarspacer, int lminbox1, int lmaxbox1, int lminbox2, int lmaxbox2, bool box2oblig, gsl_rng* r, Tree* tree);
  ~Mmodel(); 
  void MemAllocAlgo(Sequences* seqs);
  void Compute_logl_cumsum_background_seq(Sequences* seqs);
  void InitAlgo();
  int sampleProbVect(int n, double* p);
  int sampleProbVect(int n, double* v, double vsum);
  void Update_pmatbox(Sequences* seqs); 
  void Update_pvarspacer();
  void Update_pmotifs(bool update_probavect);
  void Compute_logl_ratio_box1(Sequences* seqs, int imotif);
  void Compute_logl_ratio_box2(Sequences* seqs, int imotif);
  void Compute_logl_ratio_boxes(Sequences* seqs);
  void Compute_logl_seq_by_motif();
  void Compute_logl_seq_for_motif(int imotif);
  void Update_refpos();
  void Update_Switchvarspacer();
  void Update_Sizevarspacer();
  void Update_Box2isoptional();  
  void Update_Box2Width(Sequences* seqs, int imotif);
  void Update_Box1Width(Sequences* seqs, int imotif);
  void Shift_Box1Width(Sequences* seqs, int imotif);
  void Shift_Box2Width(Sequences* seqs, int imotif);
  void Compute_ratio_seqt();
  void Compute_ratio_seqt_for_motif(int imotif);
  void Compute_logltot_and_allocimotif();
  void Alloc_tbox1refpos_and_varspacer_allseq();
  void Alloc_tbox1refpos_and_varspacer(int iseq, int imotif);
  int tbox1_from_tbox1refpos(int ref_posbox1, int tbox1refpos);
  int tbox1refpos_from_tbox1(int ref_posbox1, int tbox1);
  void printParams(ostream& os);
  void printAlloc(ostream &os, Sequences* seqs);
  void ComputeLoglRatioCurrentAlloc(double* loglbox1, double* loglbox2, int iseq);
  void simulSeqM1(Sequences* seqs);
  void Update_BackgroundMr(Sequences* seqs);
  double ComputeLogl_BackgroundMrOnly(Sequences* seqs);
  void Update_p_pos_motif(); 
  double Compute_loglik_p_pos_motif(int *breakpoints);
  double logltot();
  void set_temp(double temp);
  void heuristic_split(Sequences* seqs);
  void mergeandsplit(Sequences* seqs);
  void Alloc_counts_subtree();
  void InitCountsSubtree(int iseq, int rcomp, int compstatus, int* tseq);
  double ComputeLogLikForBoxCountVectComp1(int** vect, int dim, bool zeroflag);
  double ComputeLogLikForBoxCountVectComp2(int** vect, int** vectroot, int dim, bool zeroflag);
  double ComputeLogLikForSpacerCount(int* vect, int* vectroot, int dim, bool zeroflag);
  double ComputeLogLikForPbox2(int kbox, int knobox);

};

inline void Mmodel::set_temp(double temp)
{
  _invtemp = 1.0/temp;
}

inline double Mmodel::logltot()
{
  return(_logltot);
}

inline int Mmodel::tbox1_from_tbox1refpos(int ref_posbox1, int tbox1refpos)
{
  return(tbox1refpos-ref_posbox1);
}

inline int Mmodel::tbox1refpos_from_tbox1(int ref_posbox1, int tbox1)
{
  return(tbox1+ref_posbox1);
}

inline int Mmodel::sampleProbVect(int n, double* p) 
{
  double tx = gsl_rng_uniform(_r);
  int ri=0;
  double tsum = p[ri];
  while (tx > tsum) {
    ri++;
    tsum+=p[ri];
  }    
  if (ri>n-1) {
    ri=n-1;
  }
  return(ri);
}

inline int Mmodel::sampleProbVect(int n, double* v, double vsum) 
{
  double tx = gsl_rng_uniform(_r)*vsum;
  int ri=0;
  double tsum = v[ri];
  while (tx > tsum) {
    ri++;
    tsum+=v[ri];
  }    
  if (ri>n-1) {
    ri=n-1;
  }
  return(ri);
}

#endif /* MMODEL_H */
