// TREEMM: a software for identification of bipartite motifs in promoter sequences
// Copyright (C) 2011  Pierre NICOLAS
//
// This file is part of TREEMM.
//
//     TREEMM is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     Foobar is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with TREEMM.  If not, see <http://www.gnu.org/licenses/>.

#include <list>
#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdlib>
#include <sstream>
#include <cmath>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
using namespace std;

#include "Tree.h"

Tree::Tree(char* file, int nleaves)
{

  _nleaves = nleaves;
  _nnodes = 2*nleaves - 1;

  _nodeid = new int [_nnodes];
  _nodestatus = new int [_nnodes];
  _height = new double [_nnodes];
  _child1 = new int [_nnodes];
  _child2 = new int [_nnodes];
  _ancestor = new int [_nnodes];

  for (int i=0; i<_nnodes; i++) {
    _nodeid[i]=i;
  }

  for (int i=0; i<_nleaves; i++) {
    _height[i]=0.0;
    _child1[i]=0;
    _child2[i]=0;
    _nodestatus[i]=0;
    _ancestor[i]=-1;
  }

  ifstream ifile(file);
  string word;

  if ( ! ifile.good() )
    {
      cerr << "Cannot open the sequence list file:" << endl
	   << "\t\"" << file << "\"" << endl;
      exit(1);
    }

  for (int iheader=0; iheader<4; iheader++) {
    ifile >> word;
  }

  for (int i=_nleaves; i<_nnodes; i++) {
    ifile >> word;
    int icheck = atoi(word.c_str());
    if (i-_nleaves+1 != icheck) {
      cerr << "pb here" << endl;
      exit(0);
    }
    ifile >> word;
    _child1[i]=atoi(word.c_str());
    ifile >> word;
    _child2[i]=atoi(word.c_str());
    ifile >> word;    
    _height[i]=atof(word.c_str());
    if (_child1[i]<0) {
      _child1[i] = -_child1[i] - 1;
    } else {
      _child1[i] = _child1[i] - 1 + _nleaves;
    }
    if (_child2[i]<0) {
      _child2[i] = -_child2[i] - 1;
    } else {
      _child2[i] = _child2[i] - 1 + _nleaves;
    }
    _ancestor[_child1[i]] = i;
    _ancestor[_child2[i]] = i;
    _nodestatus[i] = 1;
    _ancestor[i]=-1;
  }
  _nodestatus[_nnodes-1] = 2;
  ifile.close();
  
  double tsmall=_height[_nnodes-1]*1e-10;
  for (int i=_nleaves; i<_nnodes; i++) {
    double t1 = _height[i] - _height[_child1[i]];
    double t2 = _height[i] - _height[_child2[i]];
    if (t1<tsmall || t2<tsmall) {
      _height[i]+=tsmall;
      cout << "Warning: branch lengths associated to need id " << _nodeid[i] << " are too small" << endl
	   << "         the height of this node has been shifted up of " << tsmall << endl; 
    }
  }
}

Tree::~Tree()
{
}

void Tree::AllocLikelihoodComputation(int nchars,gsl_rng* r, double alpha_dirich_pmotifs)
{
  _r = r;
  _nchars = nchars;
  _alpha_dirich_pmotifs = alpha_dirich_pmotifs; // Jeffrey's prior 0.5;
  _simul_id = new int [_nnodes];
  _simul_id_leaves = new int [_nleaves];
  _simul_nchanges = new int [_nnodes];
  _simul_nchanges_leaves = new int [_nleaves];
  _logl_by_char_ontree = new double* [_nnodes];
  _count_changes_subtree = new int* [_nnodes];
  _upstream_of_original_split = new bool [_nnodes];
  for (int inode=0; inode<_nnodes; inode++) {
    _count_changes_subtree[inode] = new int [_nchars];
    _logl_by_char_ontree[inode] = new double [_nchars];    
  }
  _logl_by_char_onleaves = new double* [_nleaves];
  for (int il=0; il<_nleaves; il++) {
    _logl_by_char_onleaves[il] = new double [_nchars];    
  }
  
}

double Tree::ComputeLogTotAndAllocimotif(double epsilon_evolontree, double alpha_evolontree, double *pmotifs, double *pmotifs_ontree, double** logl_seq_by_motif, int* alloc_imotif)
{
  double* tmp_x1 = new double [_nchars];
  double* tmp_x2 = new double [_nchars];
  double* log_pmotifs_ontree = new double [_nchars];
  double* log_pmotifs = new double [_nchars];
  for (int ichar=0; ichar<_nchars; ichar++) {
    log_pmotifs[ichar] = log(pmotifs[ichar]);
    log_pmotifs_ontree[ichar] = log(pmotifs_ontree[ichar]);
  }

  // initialization of bottom-up recursion
  for (int inode=0; inode<_nleaves; inode++) {
    for (int ichar=0; ichar<_nchars; ichar++) {
      _logl_by_char_onleaves[inode][ichar] = logl_seq_by_motif[inode][ichar];   
    }
  }

  // first level : changes on leaves
  double log_pnochange_onleaf = log(1.0 - epsilon_evolontree);
  double log_pchange_onleaf = log(epsilon_evolontree);
  for (int inode=0; inode<_nleaves; inode++) {
    for (int icharchild=0; icharchild<_nchars; icharchild++) {
      tmp_x1[icharchild] = log_pchange_onleaf + log_pmotifs[icharchild] + _logl_by_char_onleaves[inode][icharchild];
    }
    double log_pchange = log_sum_exp_log(_nchars, tmp_x1);
    
    for (int ichar=0; ichar<_nchars; ichar++) {
      double log_pnochange = log_pnochange_onleaf + _logl_by_char_onleaves[inode][ichar];
      _logl_by_char_ontree[inode][ichar] = log_sum2_exp_log(log_pnochange,log_pchange);
    }
  }

  // second level : changes on tree
  for (int inode=_nleaves; inode<_nnodes; inode++) {
    double t1 = _height[inode] - _height[_child1[inode]];
    double t2 = _height[inode] - _height[_child2[inode]];
    double log_pchange_child1 = log(1-exp(-alpha_evolontree*t1));
    double log_pchange_child2 = log(1-exp(-alpha_evolontree*t2));
    double log_pnochange_child1 = -alpha_evolontree*t1;
    double log_pnochange_child2 = -alpha_evolontree*t2;

    for (int icharchild=0; icharchild<_nchars; icharchild++) {
      tmp_x1[icharchild] = log_pchange_child1 + log_pmotifs_ontree[icharchild] + _logl_by_char_ontree[_child1[inode]][icharchild];
      tmp_x2[icharchild] = log_pchange_child2 + log_pmotifs_ontree[icharchild] + _logl_by_char_ontree[_child2[inode]][icharchild];
    }

    double log_p1change = log_sum_exp_log(_nchars, tmp_x1);
    double log_p2change = log_sum_exp_log(_nchars, tmp_x2);;

    for (int ichar=0; ichar<_nchars; ichar++) {
      double log_p1nochange = log_pnochange_child1 + _logl_by_char_ontree[_child1[inode]][ichar];
      double log_p2nochange = log_pnochange_child2 + _logl_by_char_ontree[_child2[inode]][ichar];
      double log_p1 = log_sum2_exp_log(log_p1change, log_p1nochange);
      double log_p2 = log_sum2_exp_log(log_p2change, log_p2nochange);
      _logl_by_char_ontree[inode][ichar] = log_p1 + log_p2;
    }
  }

  for (int ichar=0; ichar<_nchars; ichar++) {
    _logl_by_char_ontree[_nnodes-1][ichar] += log_pmotifs_ontree[ichar];
  }    
  double logltot = log_sum_exp_log(_nchars, _logl_by_char_ontree[_nnodes-1]);

  delete [] tmp_x1;
  delete [] tmp_x2;
  delete [] log_pmotifs_ontree;

  this->SimulAndAllocimotif(epsilon_evolontree,alpha_evolontree,pmotifs,pmotifs_ontree,alloc_imotif);

  return(logltot);
}

void Tree::SimulAndAllocimotif(double epsilon_evolontree, double alpha_evolontree, double *pmotifs, double *pmotifs_ontree, int* alloc_imotif)
{
  //cout << "***** SimulAndAllocimotif ***" << endl;
  double max_x;
  double sum_x;
  double* pvect = new double [_nchars+1];
  double* lvect = new double [_nchars+1];
  double* log_pmotifs_ontree = new double [_nchars];
  double* log_pmotifs = new double [_nchars];
  for (int ichar=0; ichar<_nchars; ichar++) {
    log_pmotifs[ichar] = log(pmotifs[ichar]);
    log_pmotifs_ontree[ichar] = log(pmotifs_ontree[ichar]);
  }

  // simulation of the char at root node
  sum_x = logl2prob(_nchars, _logl_by_char_ontree[_nnodes-1], pvect);
  _simul_id[_nnodes-1] = this->sampleProbVect(_nchars, pvect, sum_x);
  //cout << "root " << _simul_id[_nnodes-1] << endl;

  // backward simulation of non root nodes
  for (int inode=_nnodes-2; inode>=0; inode--) {
    double ta = _height[_ancestor[inode]] - _height[inode];
    double log_pchange_ancestor = log(1-exp(-alpha_evolontree*ta));
    double log_pnochange_ancestor = -alpha_evolontree*ta;
    lvect[_nchars] =  _logl_by_char_ontree[inode][_simul_id[_ancestor[inode]]] 
      + log_pnochange_ancestor; 
    max_x = lvect[_nchars];
    for (int ichar=0; ichar<_nchars; ichar++) {
      lvect[ichar] = _logl_by_char_ontree[inode][ichar] + log_pchange_ancestor 
	+ log_pmotifs_ontree[ichar];
    }
    sum_x = logl2prob(_nchars+1,lvect,pvect);
    _simul_id[inode] = this->sampleProbVect(_nchars+1, pvect, sum_x);

    if (_simul_id[inode]==_nchars) {
      _simul_id[inode] = _simul_id[_ancestor[inode]];
      _simul_nchanges[inode] = 0;
    } else {
      _simul_nchanges[inode] = 1;
      double tx = gsl_rng_uniform(_r)*(1 - exp(-alpha_evolontree*ta));
      double px = (alpha_evolontree*ta)*exp(-alpha_evolontree*ta);
      double sx = px;
      while (sx < tx) {
	_simul_nchanges[inode]++;
	px *= (alpha_evolontree*ta)/(double)_simul_nchanges[inode];
	sx += px;
      }
    }
  }

  double log_pnochange_onleaf = log(1.0 - epsilon_evolontree);
  double log_pchange_onleaf = log(epsilon_evolontree);
  for (int inode=0; inode<_nleaves; inode++) {
    lvect[_nchars] =  _logl_by_char_onleaves[inode][_simul_id[inode]] 
      + log_pnochange_onleaf; 
    for (int ichar=0; ichar<_nchars; ichar++) {
      lvect[ichar] = _logl_by_char_onleaves[inode][ichar] + log_pchange_onleaf 
	+ log_pmotifs[ichar];
    }
    sum_x = logl2prob(_nchars+1,lvect,pvect);
    _simul_id_leaves[inode] = this->sampleProbVect(_nchars+1, pvect, sum_x);
    if (_simul_id_leaves[inode]==_nchars) {
      _simul_id_leaves[inode] = _simul_id[inode];
      _simul_nchanges_leaves[inode]=0;
    } else {
      _simul_nchanges_leaves[inode]=1;
    }
    for (int ichar=0; ichar<_nchars; ichar++) {
      alloc_imotif[inode] = _simul_id_leaves[inode];   
    }
  }

  delete [] log_pmotifs_ontree;
  delete [] log_pmotifs;
  delete [] pvect;
  delete [] lvect;

  //  this->Update_alpha_and_pmotifs(NULL, NULL);
}

void Tree::Update_alpha_and_pmotifs(double* epsilon_evolontree, double* alpha_evolontree, double *pmotifs, double *pmotifs_ontree)
{
  double count_sum_t = 0;
  int count_sum_n = 0;
  int* count_sample = new int [_nchars];
  for (int ichar=0; ichar<_nchars; ichar++) {
    count_sample[ichar] = 0;
  }

  count_sample[_simul_id[_nnodes-1]]++;
  for (int inode=_nnodes-2; inode>=0; inode--) {
    double ta = _height[_ancestor[inode]] - _height[inode];
    count_sum_t += ta;
    if (_simul_nchanges[inode]>0) {
      count_sum_n += _simul_nchanges[inode];
      count_sample[_simul_id[inode]]++;
    }
  }

  double* alpha = new double [_nchars];
  for (int ichar=0; ichar<_nchars; ichar++) {
    alpha[ichar] = _alpha_dirich_pmotifs + (double)count_sample[ichar];
  }  
  if (pmotifs_ontree != NULL) {
    gsl_ran_dirichlet(_r, _nchars, alpha, pmotifs_ontree);
  }
  *alpha_evolontree = gsl_ran_gamma(_r,1.0+(double)count_sum_n, 1.0/(1.0 + (double)count_sum_t));


  for (int ichar=0; ichar<_nchars; ichar++) {
    count_sample[ichar] = 0;
  }
  for (int ileaf=0; ileaf<_nleaves; ileaf++) {
    if (_simul_nchanges_leaves[ileaf]==1) {
      count_sample[_simul_id_leaves[ileaf]]++;
    }    
  }
  
  count_sum_n = 0;
  for (int ichar=0; ichar<_nchars; ichar++) {
    count_sum_n += count_sample[ichar];
    alpha[ichar] = _alpha_dirich_pmotifs + (double)count_sample[ichar];
  }  
  if (pmotifs != NULL) {
    gsl_ran_dirichlet(_r, _nchars, alpha, pmotifs);
  }  
  *epsilon_evolontree = gsl_ran_beta(_r,1.0+(double)count_sum_n, 1.0+(double)(_nleaves-count_sum_n));

  delete [] alpha;
  delete [] count_sample;
}

void Tree::Update_alpha_and_pmotifs_undistinct(double* epsilon_evolontree, double* alpha_evolontree, double *pmotifs, double *pmotifs_ontree)
{
  double count_sum_t = 0;
  int count_sum_n = 0;
  int count_sum_nleaf = 0;
  int* count_sample = new int [_nchars];

  for (int ichar=0; ichar<_nchars; ichar++) {
    count_sample[ichar] = 0;
  }

  count_sample[_simul_id[_nnodes-1]]++;
  for (int inode=_nnodes-2; inode>=0; inode--) {
    double ta = _height[_ancestor[inode]] - _height[inode];
    count_sum_t += ta;
    if (_simul_nchanges[inode]>0) {
      count_sum_n += _simul_nchanges[inode];
      count_sample[_simul_id[inode]]++;
    }
  }
  for (int ileaf=0; ileaf<_nleaves; ileaf++) {
    if (_simul_nchanges_leaves[ileaf]==1) {
      count_sum_nleaf++;
      count_sample[_simul_id_leaves[ileaf]]++;
    }    
  }

  double* alpha = new double [_nchars];
  for (int ichar=0; ichar<_nchars; ichar++) {
    alpha[ichar] = _alpha_dirich_pmotifs + (double)count_sample[ichar];
  }  
  if (pmotifs != NULL) {
    gsl_ran_dirichlet(_r, _nchars, alpha, pmotifs);
    for (int ichar=0; ichar<_nchars; ichar++) {
      pmotifs_ontree[ichar] = pmotifs[ichar];
    }
  }
  *alpha_evolontree = gsl_ran_gamma(_r,1.0+(double)count_sum_n, 1.0/(1.0 + (double)count_sum_t));
  *epsilon_evolontree = gsl_ran_beta(_r,1.0+(double)count_sum_nleaf, 1.0+(double)(_nleaves-count_sum_n));

  delete [] alpha;
  delete [] count_sample;
}

void Tree::Update_alpha_and_pmotifs_undistinct_pnomotif(double* epsilon_evolontree, double* alpha_evolontree, double *pmotifs, double *pmotifs_ontree)
{
  double count_sum_t = 0;
  int count_sum_n = 0;
  int count_sum_nleaf = 0;
  int* count_sample = new int [_nchars];
  int* count_sample_leaf = new int [_nchars];

  for (int ichar=0; ichar<_nchars; ichar++) {
    count_sample[ichar] = 0;
    count_sample_leaf[ichar] = 0;
  }

  count_sample[_simul_id[_nnodes-1]]++;
  for (int inode=_nnodes-2; inode>=0; inode--) {
    double ta = _height[_ancestor[inode]] - _height[inode];
    count_sum_t += ta;
    if (_simul_nchanges[inode]>0) {
      count_sum_n += _simul_nchanges[inode];
      count_sample[_simul_id[inode]]++;
    }
  }
  for (int ileaf=0; ileaf<_nleaves; ileaf++) {
    if (_simul_nchanges_leaves[ileaf]==1) {
      count_sum_nleaf++;
      count_sample_leaf[_simul_id_leaves[ileaf]]++;
    }    
  }

  double* alpha = new double [_nchars];
  int count_sum_n_diff = 0;
  for (int ichar=0; ichar<_nchars; ichar++) {
    alpha[ichar] = _alpha_dirich_pmotifs + (double)count_sample[ichar] + (double)count_sample_leaf[ichar];
    count_sum_n_diff += count_sample[ichar];
  }  
  if (pmotifs != NULL) {
    gsl_ran_dirichlet(_r, _nchars-1, alpha, pmotifs); //nchars-1 instead of _nchars
    for (int ichar=0; ichar<_nchars-1; ichar++) {
      pmotifs_ontree[ichar] = pmotifs[ichar];
    }
    double a = 1.0 + count_sample_leaf[_nchars-1];
    double b = 1.0 + count_sum_nleaf - count_sample_leaf[_nchars-1];
    pmotifs[_nchars-1]=gsl_ran_beta(_r, a, b);    
    a = 1.0 + count_sample[_nchars-1];
    b = 1.0 + count_sum_n_diff - count_sample[_nchars-1];
    pmotifs_ontree[_nchars-1]=gsl_ran_beta(_r, a, b); 
    //cout << pmotifs_ontree[_nchars-1] << " " << a << " " << b << endl;
    for (int ichar=0; ichar<_nchars-1; ichar++) {
      pmotifs_ontree[ichar] = pmotifs_ontree[ichar]*(1-pmotifs_ontree[_nchars-1]);
      pmotifs[ichar] = pmotifs[ichar]*(1-pmotifs[_nchars-1]);
    }  
  }
  *alpha_evolontree = gsl_ran_gamma(_r,1.0+(double)count_sum_n, 1.0/(1.0 + (double)count_sum_t));
  *epsilon_evolontree = gsl_ran_beta(_r,1.0+(double)count_sum_nleaf, 1.0+(double)(_nleaves-count_sum_n));

  delete [] alpha;
  delete [] count_sample;
  delete [] count_sample_leaf;
}

void Tree::IndentifyCandidateForMergeAndSplit(int** compat_pairs)
{
  // We search for pairs of motifs such that motifs segregate 
  // in the node of the tree 
  // according to a branch of the tree.

  int* count_changes_tot = new int [_nchars];
  int* count_states_leaf = new int [_nchars];

  for (int ichar=0; ichar<_nchars; ichar++) {
    count_changes_tot[ichar] = 0;
    count_states_leaf[ichar] = 0;
  }

  count_changes_tot[_simul_id[_nnodes-1]]++;
  for (int inode=0; inode<_nnodes-1; inode++) {
    if (_simul_nchanges[inode]>0) {
      count_changes_tot[_simul_id[inode]]++;
    }
  }
  for (int ileaf=0; ileaf<_nleaves; ileaf++) {
    count_states_leaf[_simul_id_leaves[ileaf]]++;
    if (_simul_nchanges_leaves[ileaf]==1) {
      count_changes_tot[_simul_id_leaves[ileaf]]++;
    }  
  }

  int** count_changes_subtree = new int* [_nnodes];

  for (int inode=0; inode<_nnodes; inode++) {
    count_changes_subtree[inode] = new int [_nchars];
  }

  for (int ic=0; ic<_nchars; ic++) {
    for (int jc=0; jc<_nchars; jc++) {
      compat_pairs[ic][jc] = 0;
    }  
  }  

  for (int inode=0; inode<_nnodes; inode++) {
    if (_nodestatus[inode]==0) {
      for (int ic=0; ic<_nchars; ic++) {
	count_changes_subtree[inode][ic] = 0;
      }
      if (_simul_nchanges_leaves[inode]==1) {
	count_changes_subtree[inode][_simul_id_leaves[inode]]++;
      }    
    } else {
      for (int ic=0; ic<_nchars; ic++) {
	count_changes_subtree[inode][ic] 
	  = count_changes_subtree[_child1[inode]][ic] 
	  + count_changes_subtree[_child2[inode]][ic];
      }      
    }    
    if (_simul_nchanges[inode]>0 || _nodestatus[inode]==2) {
      count_changes_subtree[inode][_simul_id[inode]]++;
    }
  }    

  int* tmotifs_all = new int [_nchars];
  int* tmotifs_0 = new int [_nchars];
  for (int inode=0; inode<_nnodes; inode++) {
    int n_0=0;
    int n_all=0;
    for (int ic=0; ic<_nchars; ic++) {
      if (count_changes_subtree[inode][ic]==count_changes_tot[ic]) {
	tmotifs_all[n_all]=ic;
	n_all++;
      } 
      if (count_changes_subtree[inode][ic]==0 && _simul_id[inode] != ic) {
	tmotifs_0[n_0]=ic;
	n_0++;
      }
    }
    for (int iall=0; iall<n_all; iall++) {
      for (int i0=0; i0<n_0; i0++) {
	compat_pairs[tmotifs_all[iall]][tmotifs_0[i0]]++;
      }  
    }
  }
  delete [] tmotifs_all;
  delete [] tmotifs_0;

  if (0) {
    cout << "count_states_leaf" << endl;
    for (int ic=0; ic<_nchars; ic++) {
      cout << " " << ic << ":" << count_states_leaf[ic];
    }  
    cout << endl;
    cout << "count_changes_tot" << endl;
    for (int ic=0; ic<_nchars; ic++) {
      cout << " " << ic << ":" << count_changes_tot[ic];
    }  
    cout << endl;

    cout << "compat_pairs" << endl;
    for (int ic=0; ic<_nchars; ic++) {
      for (int jc=0; jc<_nchars; jc++) {
	if (ic==jc) {
	  cout << " " << "x";
	} else if (compat_pairs[ic][jc]==0) {
	  cout << " " << ".";
	} else {
	  cout << " " << compat_pairs[ic][jc];
	}
      }  
      cout << endl;
    }  
  }

  for (int ic=0; ic<_nchars; ic++) {
    for (int jc=0; jc<ic; jc++) {
      compat_pairs[ic][jc]+=compat_pairs[jc][ic];
      compat_pairs[jc][ic]=0;
      if (compat_pairs[ic][jc]>0) {
	compat_pairs[ic][jc]=count_changes_tot[ic]+count_changes_tot[jc];
      }
    }
  }  

  for (int inode=0; inode<_nnodes; inode++) {
    delete [] count_changes_subtree[inode];
  }
  delete [] count_changes_subtree;
  delete [] count_changes_tot;
  delete [] count_states_leaf;
}

void Tree::CompatPairsForMergeAndSplit(int** compat_pairs, int* &simul_nchanges_leaves)
{
  // We search for pairs of motifs such that motifs segregate 
  // in the node of the tree 
  // according to a branch of the tree.

  simul_nchanges_leaves = _simul_nchanges_leaves;

  for (int ic=0; ic<_nchars; ic++) {
    for (int jc=0; jc<_nchars; jc++) {
      compat_pairs[ic][jc] = -1;
    }  
  }  

  for (int inode=0; inode<_nnodes; inode++) {
    if (_nodestatus[inode]==0) {
      for (int ic=0; ic<_nchars; ic++) {
	_count_changes_subtree[inode][ic] = 0;
      }
//       if (_simul_nchanges_leaves[inode]==1) {
// 	_count_changes_subtree[inode][_simul_id_leaves[inode]]++;
//       }    
    } else {
      for (int ic=0; ic<_nchars; ic++) {
	_count_changes_subtree[inode][ic] 
	  = _count_changes_subtree[_child1[inode]][ic] 
	  + _count_changes_subtree[_child2[inode]][ic];
      }      
    }    
    if (_simul_nchanges[inode]>0 || _nodestatus[inode]==2) {
      _count_changes_subtree[inode][_simul_id[inode]]++;
    }
  }    

  int* tmotifs_all = new int [_nchars];
  int* tmotifs_0 = new int [_nchars];
  for (int inode=0; inode<_nnodes; inode++) {
    int n_0=0;
    int n_all=0;
    for (int ic=0; ic<_nchars; ic++) {
      if (_count_changes_subtree[_nnodes-1][ic]>0) {
	if (_count_changes_subtree[inode][ic]==_count_changes_subtree[_nnodes-1][ic]
	    && _count_changes_subtree[inode][ic]!=_count_changes_subtree[_child1[inode]][ic]
	    && _count_changes_subtree[inode][ic]!=_count_changes_subtree[_child2[inode]][ic]) {
	  tmotifs_all[n_all]=ic;
	  n_all++;
	} 
	if (_count_changes_subtree[inode][ic]==0 && _simul_id[inode] != ic) {
	  tmotifs_0[n_0]=ic;
	  n_0++;
	}
      }
    }
    for (int iall=0; iall<n_all; iall++) {
      for (int i0=0; i0<n_0; i0++) {
	compat_pairs[tmotifs_all[iall]][tmotifs_0[i0]] = inode;
      }  
    }
  }
  for (int ic=0; ic<_nchars; ic++) {
    if (_count_changes_subtree[_nnodes-1][ic]==0) {
      for (int jc=0; jc<_nchars; jc++) {
	if (_count_changes_subtree[_nnodes-1][jc]>0) {
	  compat_pairs[ic][jc]=_nnodes-1;
	}	
      }      
    }
  }
  for (int ic=0; ic<_nchars; ic++) {
    for (int jc=0; jc<_nchars; jc++) {
      if (compat_pairs[ic][jc]>=0 && compat_pairs[ic][jc]!=_nnodes-1) {
	if (_simul_id[compat_pairs[ic][jc]]==ic 
	    && _simul_id[_ancestor[compat_pairs[ic][jc]]]==jc 
	    && _simul_nchanges[compat_pairs[ic][jc]] > 1) {
	  compat_pairs[ic][jc]=-1;
	  // as in this case in the forward move _simul_nchanges[compat_pairs[ic][jc]] will be set to 0
	  // and in the reverse move _simul_nchanges[compat_pairs[ic][jc]] will be set back to 1
	}
      }
    }
  }
  delete [] tmotifs_all;
  delete [] tmotifs_0;
}

void Tree::printTree()
{
  cout << "***** printTree" << endl;
  cout << "inode" << "\t" << "status" << "\t" << "child1" << "\t" << "child2" << "\t" << "id" << "\t" << "nch" << "\t" << "idleav" << "\t" << "nchleav" << endl;
  for (int inode=0; inode<_nnodes; inode++) {
    cout << inode << "\t" << _nodestatus[inode] << "\t" << _child1[inode] << "\t" << _child2[inode] << "\t" << _simul_id[inode] << "\t" << _simul_nchanges[inode];
    if (_nodestatus[inode]==0) {
      cout << "\t" << _simul_id_leaves[inode] << "\t" << _simul_nchanges_leaves[inode];
    }
    cout << endl;
  }
  cout << "*****" << endl;
}

void Tree::recursSubtree(int lmaxbox1, int lmaxbox2, int lmaxvarspacer, int rcomp1, int rcomp2,
			 double alpha_evolontree, double* pmotifs_ontree,
			 int**** count_nuc_matbox1_subtree, 
			 int**** count_nuc_matbox2_subtree,
			 int** count_varspacer_subtree,
			 int** count_box2nobox2_subtree, 
			 int* same_count_as_subtree, bool* allowedsplit_subtree,
			 double** loglikbackground_subtree,
			 double* liktreesplit_subtree, int original_splitnode)
{
  int* count_nnodes_subtree = new int [_nnodes];
  //cout << "inode" << "\t" << "_child1" << " " << "_child2" << " " << "allowedsplit" << "\t" << "count_nnodes_subtree" << "\t" << "same_count_as_subtree" << " " << "loglikbackground_subtree" << " " << "loglikbackground_subtree" << endl;
  bool special_case = (original_splitnode != _nnodes-1);
  if (special_case) {
    special_case = (_simul_nchanges[original_splitnode]==1 
		    && _simul_id[_ancestor[original_splitnode]]==rcomp2);
    // in which case this change is removed in the forward move is the new split node != original_splitnode
  }
  cout << "_count_changes_subtree[_nnodes-1]";
  for (int ic=0; ic<_nchars; ic++) {
    cout << " " << ic << ":" << _count_changes_subtree[_nnodes-1][ic];
  }
  cout << endl;
  cout << "special case " << special_case << endl;
  for (int inode=0; inode<_nnodes; inode++) {
    if (_nodestatus[inode]==0) {
      _upstream_of_original_split[inode]=false;
      count_nnodes_subtree[inode]=0;
//       if (_simul_id_leaves[inode]==rcomp1 || _simul_id_leaves[inode]==rcomp2) {
// 	count_nnodes_subtree[inode]++;
//       }
      if (_simul_id[inode]==rcomp1 || _simul_id[inode]==rcomp2) {
	count_nnodes_subtree[inode]++;	
      }
    } else {
      _upstream_of_original_split[inode]=(_upstream_of_original_split[_child1[inode]] || _upstream_of_original_split[_child2[inode]]);
      for (int ic=0; ic<_nchars; ic++) {
	count_nnodes_subtree[inode]
	  = count_nnodes_subtree[_child1[inode]] 
	  + count_nnodes_subtree[_child2[inode]];
      }    
      if (_simul_id[inode]==rcomp1 || _simul_id[inode]==rcomp2) {
	count_nnodes_subtree[inode]++;	
      }  
    }
    if (inode==original_splitnode) {
      _upstream_of_original_split[inode]=true;
    }
    bool allowedsplit = true;
    if (_nodestatus[inode]!=0) {
      same_count_as_subtree[inode] = inode;
    }
    if (count_nnodes_subtree[inode]==0) {
      allowedsplit = false;
      if (_nodestatus[inode]!=0) {
	same_count_as_subtree[inode]=-1;
      }
    } 
    if (_nodestatus[inode]!=0) { 
      if (count_nnodes_subtree[inode]==count_nnodes_subtree[_child1[inode]]) {
	allowedsplit = false;
	same_count_as_subtree[inode] = same_count_as_subtree[_child1[inode]];
      } else if (count_nnodes_subtree[inode]==count_nnodes_subtree[_child2[inode]]) {
	allowedsplit = false;
	same_count_as_subtree[inode] = same_count_as_subtree[_child2[inode]];
      } 
      if (_nodestatus[inode]==2) {
	allowedsplit = true; // the "split" at the root has a special meaning. it means changing all rcomp1 to rcomp2.
      }
    }
    if (_nodestatus[inode]!=2) { 
      if ((_simul_id[inode]==rcomp1 || _simul_id[inode]==rcomp2) && (_simul_nchanges[inode]>0) 
	  && (_simul_id[inode]==_simul_id[_ancestor[inode]]) ) {
	allowedsplit = false;
      }    
    }
    if (_nodestatus[inode]!=2) { 
      if (_count_changes_subtree[inode][rcomp2]==_count_changes_subtree[_nnodes-1][rcomp2]) {
	allowedsplit = false;
	// this would means changing all rcomp2 to rcomp1
      }
    }    
    if (_nodestatus[inode]!=0 && same_count_as_subtree[inode]==inode) {
      for (int i=0; i<2; i++) {
	int tsum = 0;
	if (same_count_as_subtree[_child1[inode]]>=0) {
	  tsum += count_box2nobox2_subtree[same_count_as_subtree[_child1[inode]]][i];
	}
	if (same_count_as_subtree[_child2[inode]]>=0) {
	  tsum += count_box2nobox2_subtree[same_count_as_subtree[_child2[inode]]][i];
	}
	count_box2nobox2_subtree[inode][i] = tsum;
      }
      for (int ivs=0; ivs<lmaxvarspacer; ivs++) {
	int tsum = 0;
	if (same_count_as_subtree[_child1[inode]]>=0) {
	  tsum += count_varspacer_subtree[same_count_as_subtree[_child1[inode]]][ivs];
	}
	if (same_count_as_subtree[_child2[inode]]>=0) {
	  tsum += count_varspacer_subtree[same_count_as_subtree[_child2[inode]]][ivs];
	}
	count_varspacer_subtree[inode][ivs] = tsum;
      }
      for (int im=0; im<2; im++) {
	double tsum = 0;
	if (same_count_as_subtree[_child1[inode]]>=0) {
	  tsum += loglikbackground_subtree[same_count_as_subtree[_child1[inode]]][im];
	}
	if (same_count_as_subtree[_child2[inode]]>=0) {
	  tsum += loglikbackground_subtree[same_count_as_subtree[_child2[inode]]][im];
	}
	loglikbackground_subtree[inode][im] = tsum;
      }
      for (int im=0; im<2; im++) {
	for (int ipos=0; ipos<lmaxbox1; ipos++) {
	  for (int ib=0; ib<4; ib++) {
	    int tsum1 = 0;
	    if (same_count_as_subtree[_child1[inode]]>=0) {
	      tsum1 += count_nuc_matbox1_subtree[same_count_as_subtree[_child1[inode]]][im][ipos][ib];
	    }
	    if (same_count_as_subtree[_child2[inode]]>=0) {
	      tsum1 += count_nuc_matbox1_subtree[same_count_as_subtree[_child2[inode]]][im][ipos][ib];
	    }
	    count_nuc_matbox1_subtree[inode][im][ipos][ib] = tsum1;
	  }
	}
	for (int ipos=0; ipos<lmaxbox2; ipos++) {
	  for (int ib=0; ib<4; ib++) {
	    int tsum2 = 0;
	    if (same_count_as_subtree[_child1[inode]]>=0) {
	      tsum2 += count_nuc_matbox2_subtree[same_count_as_subtree[_child1[inode]]][im][ipos][ib];
	    }
	    if (same_count_as_subtree[_child2[inode]]>=0) {
	      tsum2 += count_nuc_matbox2_subtree[same_count_as_subtree[_child2[inode]]][im][ipos][ib];
	    }
	    count_nuc_matbox2_subtree[inode][im][ipos][ib] = tsum2;
	  }
	}
      }
    }
    if (allowedsplit) {
      //cout << inode << "\t" << _child1[inode] << " " << _child2[inode] << " " << allowedsplit << "\t" << count_nnodes_subtree[inode] << "\t" << same_count_as_subtree[same_count_as_subtree[inode] << " " << loglikbackground_subtree[same_count_as_subtree[inode]][0] << " " << loglikbackground_subtree[same_count_as_subtree[inode]][1] << endl;
      allowedsplit_subtree[inode]=true;
      int ksum = 0;
      double sll = 0;
      int krcomp1 = 0;
      int krcomp2 = 0;
      for (int ic=0; ic<_nchars; ic++) {
	if (ic!=rcomp1 && ic!=rcomp2) {
	  int k = _count_changes_subtree[_nnodes-1][ic];
	  ksum += k;
	  //sll += gsl_sf_lngamma(1.0+(double)k) - gsl_sf_lngamma(1.0);
	} else if (ic==rcomp1) {
	  if (_nodestatus[inode]!=2) {
	    int k = _count_changes_subtree[inode][rcomp1]+_count_changes_subtree[inode][rcomp2];
	    if (_simul_nchanges[inode]==0 && (_simul_id[inode]==rcomp1 || _simul_id[inode]==rcomp2)) {
	      k++;
	      sll += log(alpha_evolontree*(_height[_ancestor[inode]]-_height[inode])) + log(pmotifs_ontree[rcomp1]);
	    }
	    if (special_case && _upstream_of_original_split[inode]==true && inode!=original_splitnode) {
	      k--;
	      //sll += -log(alpha_evolontree*(_height[_ancestor[original_splitnode]]-_height[original_splitnode]));
	      sll -= log(alpha_evolontree*(_height[_ancestor[original_splitnode]]-_height[original_splitnode])) + log(pmotifs_ontree[rcomp1]);
	    }
	    ksum += k;
	    //sll += gsl_sf_lngamma(1.0+(double)k) - gsl_sf_lngamma(1.0);	
	    krcomp1 = k;
	  }
	} else if (ic==rcomp2) {
	  int k = _count_changes_subtree[_nnodes-1][rcomp1]+_count_changes_subtree[_nnodes-1][rcomp2];
	  if (_nodestatus[inode]!=2) {	    
	    k -= (_count_changes_subtree[inode][rcomp1]+_count_changes_subtree[inode][rcomp2]);
	  }
	  if (special_case && _upstream_of_original_split[inode]==false && inode!=original_splitnode) {
	      k--;
	      //sll += -log(alpha_evolontree*(_height[_ancestor[original_splitnode]]-_height[original_splitnode]));
	      sll -= log(alpha_evolontree*(_height[_ancestor[original_splitnode]]-_height[original_splitnode])) + log(pmotifs_ontree[rcomp1]);
	  }
	  ksum += k;
	  //sll += gsl_sf_lngamma(1.0+(double)k) - gsl_sf_lngamma(1.0);	  
	  krcomp2 = k;
	}
      }
      cout << "inode " << inode << " upof " << _upstream_of_original_split[inode] << " sid " << _simul_id[inode] << " nch " << _simul_nchanges[inode];
      if (_nodestatus[inode]==0) {
	cout << " leaf " << _simul_id_leaves[inode] << " leafchg " << _simul_nchanges_leaves[inode];
      }
      if (_nodestatus[inode]!=2) {
	cout << " ances " << _simul_id[_ancestor[inode]];
      }
      cout << " nchanges " << krcomp1 << " " << krcomp2 << " " << ksum << endl;
      //sll -= gsl_sf_lngamma((double)(_nchars+ksum)) - gsl_sf_lngamma((int)_nchars);
      liktreesplit_subtree[inode]=sll;
    } else {
      allowedsplit_subtree[inode]=false;
      liktreesplit_subtree[inode]=0;
    }
  }
  delete [] count_nnodes_subtree;
}
