// TREEMM: a software for identification of bipartite motifs in promoter sequences
// Copyright (C) 2011  Pierre NICOLAS
//
// This file is part of TREEMM.
//
//     TREEMM is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     Foobar is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with TREEMM.  If not, see <http://www.gnu.org/licenses/>.

#ifndef TREE_H
#define TREE_H

#include <gsl/gsl_rng.h>
#include <string>
#include <cmath>
using namespace std;


class Tree
{
private:
  
  gsl_rng* _r;

  int _nleaves;
  int _nnodes;
  int* _nodeid; // 0 -> 
  int* _nodestatus; // 0 if leaf, 1 if interior node, 2 if root
  double* _height; // 0 if leaf
  int* _child1;
  int* _child2;
  int* _ancestor;
  
  int _nchars;
  double** _logl_by_char_onleaves;
  double** _logl_by_char_ontree;
  int* _simul_id;
  int* _simul_id_leaves;
  int* _simul_nchanges;
  int* _simul_nchanges_leaves;

  int** _count_changes_subtree;
  bool* _upstream_of_original_split;
  double _alpha_dirich_pmotifs;
  
public:
  Tree(char* file, int nleaves);
  ~Tree(); 
  void AllocLikelihoodComputation(int nchars, gsl_rng* r, double alpha_dirich_pmotifs);

  double ComputeLogTotAndAllocimotif(double epsilon_evolontree, double alpha_evolontree, double *pmotifs, double *pmotifs_ontree, double** logl_seq_by_motif, int* alloc_imotif);

  int sampleProbVect(int n, double* v, double vsum);

  void Update_alpha_and_pmotifs(double* epsilon_evolontree, double* alpha_evolontree, double *pmotifs, double *pmotifs_ontree);

  void Update_alpha_and_pmotifs_undistinct(double* epsilon_evolontree, double* alpha_evolontree, double *pmotifs, double *pmotifs_ontree);

  void Update_alpha_and_pmotifs_undistinct_pnomotif(double* epsilon_evolontree, double* alpha_evolontree, double *pmotifs, double *pmotifs_ontree);

  void SimulAndAllocimotif(double epsilon_evolontree, double alpha_evolontree, double *pmotifs, double *pmotifs_ontree, int* alloc_imotif);

  void IndentifyCandidateForMergeAndSplit(int** compat_pairs);
  void CompatPairsForMergeAndSplit(int** compat_pairs, int* &simul_nchanges_leaves);

  double log_sum_exp_log(int n, double *);
  double log_sum2_exp_log(double x1, double x2);
  double logl2prob(int n, double *v, double *p);

  void recursSubtree(int lmaxbox1, int lmaxbox2, int lmaxvarspacer, int rcomp1, int rcomp2, 
		     double alpha_evolontree, double* pmotifs_ontree,
		     int**** count_nuc_matbox1_subtree, 
		     int**** count_nuc_matbox2_subtree,
		     int** count_varspacer_subtree,
		     int** count_box2nobox2_subtree, 
		     int* same_count_as_subtree, bool* allowedsplit_subtree,
		     double** loglikbackground_subtree,
		     double* liktreesplit_subtree, int original_splitnode);
  
  void printTree();
};

inline double Tree::logl2prob(int n, double *v, double *p)
{
  double max_v = v[0];
  for (int i=1; i<n; i++) {
    if (v[i]>max_v) {
      max_v = v[i];
    }
  }
  double sum_v = 0;
  for (int i=0; i<n; i++) {
    p[i] = exp(v[i] - max_v);
    sum_v += p[i];
  }
  return(sum_v);
}

inline double Tree::log_sum_exp_log(int n, double *v)
{
  double max_v = v[0];
  for (int i=1; i<n; i++) {
    if (v[i]>max_v) {
      max_v = v[i];
    }
  }
  double sum_v = 0;
  for (int i=0; i<n; i++) {
    sum_v += exp(v[i] - max_v);
  }
  return(max_v + log(sum_v));     
}

inline double Tree::log_sum2_exp_log(double x1, double x2)
{
  double x;
  if (x1>x2) {
    x = x1 + log(1+exp(x2-x1));
  } else {
    x = x2 + log(1+exp(x1-x2));
  }
  return(x);     
}

inline int Tree::sampleProbVect(int n, double* v, double vsum) 
{
  double tx = gsl_rng_uniform(_r)*vsum;
  int ri=0;
  double tsum = v[ri];
  while (tx > tsum) {
    ri++;
    tsum+=v[ri];
  }    
  if (ri>n-1) {
    ri=n-1;
  }
  return(ri);
}

#endif /* TREE_H */
